# -*- coding: utf8 -*-

#-----------------------------------------
# This is a part of Pylote project.
# Name:         utils_instruments.py
# Author:       Pascal Peter
# Copyright:    (C) 2009-2014 Pascal Peter
# Licence:      GNU General Public Licence version 3
# Website:      http://pascal.peter.free.fr
# Email:        pascal.peter at free.fr
#-----------------------------------------

"""
DESCRIPTION :
    Ce module contient les variables et fonctions utiles au programme.
    Certaines parties spécialisées sont dans les fichiers utils_aaa.py
"""

#importation des modules utiles:
import sys
import os

import utils

if utils.PYSIDE:
    from PySide import QtCore, QtGui, QtSvg
else:
    from PyQt4 import QtCore, QtGui, QtSvg



###########################################################"
###########################################################"
#   VARIABLES GLOBALES
###########################################################"
###########################################################"

# MAX_ZVALUE pour placer l'instrument sélectionné au premier plan
MAX_ZVALUE = 1
# IS_TRACING_ENABLED pour tracer avec le compas
IS_TRACING_ENABLED = False
# WITH_FALSE_CURSOR pour savoir si le "faux" curseur est actif
WITH_FALSE_CURSOR = False

# POINT_NAMES est la liste des noms disponibles pour les points
# elle est construite à partir de LETTERS
LETTERS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
POINT_NAMES = []
for letter in LETTERS:
    POINT_NAMES.append(letter)


def changeMaxZValue(newValue):
    # pour modifier 
    global MAX_ZVALUE
    MAX_ZVALUE = newValue

def changeTracingEnabled(newValue):
    # pour modifier IS_TRACING_ENABLED
    global IS_TRACING_ENABLED
    IS_TRACING_ENABLED = newValue

def changeFalseCursor(newValue):
    # pour modifier WITH_FALSE_CURSOR
    global WITH_FALSE_CURSOR
    WITH_FALSE_CURSOR = newValue

def changePointNames(value):
    # pour retirer une lettre de la liste POINT_NAMES
    global POINT_NAMES
    if value in POINT_NAMES:
        POINT_NAMES.pop(POINT_NAMES.index(value))








###########################################################"
###########################################################"
#                LES INSTRUMENTS DE GÉOMÉTRIE
###########################################################"
###########################################################"

class Instrument(QtSvg.QGraphicsSvgItem):
    """
    La classe de base pour un instrument de géométrie
    Les vrais instruments en dérivent
    """
    def __init__(self, parent, imageFile):
        QtSvg.QGraphicsSvgItem.__init__(self, imageFile)
        self.main = parent
        """
        MAX_ZVALUE sert à faire passer l'instrument au premier plan à la sélection
        imageFile passé en paramètre est l'image de l'instrument à afficher
        isUsed si l'instrument est en cours d'utilisation
                    pour désactiver le tracé du crayon dans ce cas
        leftButtonMode pour le mode TBI (souris un bouton) ou pas
        transformOriginPoint() est le centre pour les rotations (redéfinit pour chaque instrument)
        tracePoint est le point pour les trace (voir le compas)
        locked si l'instrument est bloqué (pour TBI ; False par défaut)
        units si l'instrument a des graduations (règle et équerre)
        """
        self.setFlags(QtGui.QGraphicsItem.ItemIsSelectable|
                        QtGui.QGraphicsItem.ItemIsMovable)
        changeMaxZValue(MAX_ZVALUE + 1)
        self.setZValue(MAX_ZVALUE)

        self.tracePoint = QtCore.QPointF(0, 0)
        self.isUsed = False
        self.leftButtonMode = 'MOVE'
        self.setCursor(QtCore.Qt.ArrowCursor)
        self.locked = False
        self.units = False
        self.A = None
        self.B = None

    def doLeftButtonMode(self, newMode):
        """
        redéfinition des Flags en fonction du mode sélectionné
        pour le bouton gauche de la souris
        """
        self.leftButtonMode = newMode
        if newMode == 'MOVE':
            self.setFlags(QtGui.QGraphicsItem.ItemIsSelectable|
                            QtGui.QGraphicsItem.ItemIsMovable)
        else:
            self.setFlags(QtGui.QGraphicsItem.ItemIsSelectable)

    def mousePressEvent(self, event):
        """
        on modifie MAX_ZVALUE pour faire passer l'instrument au premier plan
        on calcule initialLength (distance de la souris au centre) et initialAngle
        isUsed est mis à True
        """
        if self.locked:
            return
        QtGui.QGraphicsItem.mousePressEvent(self, event)
        changeMaxZValue(MAX_ZVALUE + 1)
        self.setZValue(MAX_ZVALUE)

        rayon = QtCore.QLineF(self.transformOriginPoint(), event.pos())
        self.initialLength = rayon.length()
        self.initialAngle = rayon.angle()

        if event.button() == QtCore.Qt.LeftButton:
            event.accept()
        self.isUsed = True

    def mouseMoveEvent(self, event):
        """
        on calcule length (distance de la souris au centre) et angle
        Puis suivant le mode (bouton de souris ou leftButtonMode), on fait ce qu'il faut
        """
        if self.locked:
            return
        QtGui.QGraphicsItem.mouseMoveEvent(self, event)
        rayon = QtCore.QLineF(self.transformOriginPoint(), event.pos())
        length = rayon.length()
        angle = rayon.angle()

        if event.buttons() == QtCore.Qt.RightButton:
            self.setRotation(self.rotation() + self.initialAngle - angle)
            event.accept()
        elif event.buttons() == QtCore.Qt.LeftButton:
            if self.leftButtonMode == 'MOVE':
                pass
            elif self.leftButtonMode == 'SCALE':
                if self.units and self.main.toolsWindow.actionUnitsLock.isChecked():
                    pass
                else:
                    coeff = length / self.initialLength
                    self.setScale(self.scale() * coeff)
            elif self.leftButtonMode == 'ROTATE':
                self.setRotation(self.rotation() + self.initialAngle - angle)
                event.accept()
        else:
            if self.units and self.main.toolsWindow.actionUnitsLock.isChecked():
                pass
            else:
                coeff = length / self.initialLength
                self.setScale(self.scale() * coeff)

    def mouseReleaseEvent(self, event):
        """
        isUsed est mis à False : on cesse d'utiliser l'instrument
        """
        if self.locked:
            return
        QtGui.QGraphicsItem.mouseReleaseEvent(self, event)
        if self.units and not(self.main.toolsWindow.actionUnitsLock.isChecked()):
            for instrument in self.main.listeInstruments:
                if instrument.units and not(instrument.isUsed):
                    instrument.setScale(self.scale())
        self.isUsed = False

class Protractor(Instrument):
    """
    Le Rapporteur. L'image doit être décalée pour que le centre du rapporteur soit en (0;0)
    transformOriginPoint() est donc redéfini ainsi
    """
    def __init__(self, parent):
        Instrument.__init__(self, parent, './images/instruments/rapporteur.svg')
        x, y = self.boundingRect().width() // 2, self.boundingRect().height()
        self.setTransformOriginPoint(x, y)
        self.moveBy(- x, - y)
        self.setScale(0.75)

class Ruler(Instrument):
    """
    La règle. La graduation 0 doit être l'origine des rotations.
    transformOriginPoint() est donc redéfini ainsi
    """
    def __init__(self, parent):
        Instrument.__init__(self, parent, './images/instruments/regle.svg')
        x = 20.4
        self.setTransformOriginPoint(x, 0)
        self.moveBy(- x, 0)
        settings = QtCore.QSettings(utils.PROGNAME, 'config')
        scale = float(settings.value('instrumentsScale', 1))
        self.setScale(scale)
        self.units = True
        self.A = QtCore.QPointF(x, 0)
        self.B = QtCore.QPointF(730, 0)

class Square(Instrument):
    """
    L'équerre graduée.
    """
    def __init__(self, parent):
        Instrument.__init__(self, parent, './images/instruments/equerre.svg')
        settings = QtCore.QSettings(utils.PROGNAME, 'config')
        scale = float(settings.value('instrumentsScale', 1))
        self.setScale(scale)
        self.units = True
        self.A = QtCore.QPointF(10, 0)
        self.B = QtCore.QPointF(470, 0)

class SquareNotGraduated(Instrument):
    """
    L'équerre graduée.
    """
    def __init__(self, parent):
        Instrument.__init__(self, parent, './images/instruments/equerre_non_graduee.svg')
        settings = QtCore.QSettings(utils.PROGNAME, 'config')
        self.A = QtCore.QPointF(10, 0)
        self.B = QtCore.QPointF(470, 0)

class Compass(Instrument):
    """
    Le compas doit pouvoir tracer des arcs si la variable globale IS_TRACING_ENABLED est à True.
    En fait on a besoin de 2 variables :
    
    IS_TRACING_ENABLED (variable globale) est à True quand la trace du compas est sélectionnée
        (outil sélectionné ou touche control maintenue)
    isTracingPossible est à True quand le compas est autorisé à tracer
        (bon bouton de la souris, etc...)
    isTracing est à True quand le compas est en train de tracer
    
    On redéfinit donc les évènements souris
    transformOriginPoint() est redéfini sur la pointe du compas
    tracePoint est sur la mine
    """
    def __init__(self, parent):
        Instrument.__init__(self, parent, './images/instruments/compas.svg')
        x, y = self.boundingRect().width(), self.boundingRect().height()
        self.setTransformOriginPoint(0, y)
        self.moveBy(0, - y)
        #self.setScale(0.5)
        self.tracePoint = QtCore.QPointF(x, y)
        self.isTracingPossible = False
        self.isTracing = False

    def mousePressEvent(self, event):
        """
        On commence un nouvel arc
        """
        if self.locked:
            return
        Instrument.mousePressEvent(self, event)
        self.isTracingPossible = False
        if event.buttons() == QtCore.Qt.RightButton:
            self.isTracingPossible = True
        elif event.buttons() == QtCore.Qt.LeftButton:
            if self.leftButtonMode == 'ROTATE':
                self.isTracingPossible = True
        if self.isTracingPossible and IS_TRACING_ENABLED:
            self.main.view.last = None
            center = self.mapToScene(self.transformOriginPoint())
            point = self.mapToScene(self.tracePoint)
            self.main.view.initArc(center, point)
            self.isTracing = True

    def mouseMoveEvent(self, event):
        """
        On affiche l'arc
        """
        if self.locked:
            return
        Instrument.mouseMoveEvent(self, event)
        if self.isTracing:
            self.main.view.last = None
            center = self.mapToScene(self.transformOriginPoint())
            point = self.mapToScene(self.tracePoint)
            self.main.view.drawArcTo(center, point)
            if not IS_TRACING_ENABLED:
                self.isTracing = False
        else:
            if self.isTracingPossible and IS_TRACING_ENABLED:
                self.main.view.last = None
                center = self.mapToScene(self.transformOriginPoint())
                point = self.mapToScene(self.tracePoint)
                self.main.view.initArc(center, point)
                self.isTracing = True

    def mouseReleaseEvent(self, event):
        """
        On termine l'arc
        """
        if self.locked:
            return
        Instrument.mouseReleaseEvent(self, event)
        if self.isTracing:
            if IS_TRACING_ENABLED:
                self.main.view.last = None
                center = self.mapToScene(self.transformOriginPoint())
                point = self.mapToScene(self.tracePoint)
                item = self.main.view.drawArcTo(center, point)
                self.isTracing = False
                self.main.view.drawId += 1
                item.setData(QtCore.Qt.UserRole, self.main.view.drawId)






###########################################################"
###########################################################"

#                LE FAUX CURSEUR

###########################################################"
###########################################################"

class MyCursor(QtGui.QGraphicsPixmapItem):
    """
    Le "faux" curseur qui suit la souris (utile pour certains TBI ?).
    Il est dans la liste des instruments, d'où le leftButtonMode 
        pour éviter les erreurs mais qui n'a pas d'utilité.
        (on n'a pas à le redimensionner ou le faire tourner)
    """
    def __init__(self, parent):
        QtGui.QGraphicsPixmapItem.__init__(self)
        self.main = parent

        newimage = QtGui.QImage('./images/cursor.png')
        self.setPixmap(QtGui.QPixmap.fromImage(newimage))
        
        centralPoint = QtCore.QPointF(-8, -8)
        self.setOffset(centralPoint)

        self.isUsed = False
        self.units = False
        self.leftButtonMode = 'MOVE'

    def doLeftButtonMode(self, newMode):
        """
        juste pour éviter les erreurs
        """
        self.leftButtonMode = newMode










###########################################################"
###########################################################"

#                LES POINTS CRÉÉS PAR L'UTILISATEUR

###########################################################"
###########################################################"

class PointItem(QtGui.QGraphicsPathItem):
    """
    Un point et son label.
    C'est un QGraphicsPathItem qui contient le dessin du point (path)
        et un label (QGraphicsTextItem)
    """
    def __init__(self, parent, pen, brush, font, text = ''):
        QtGui.QGraphicsPathItem.__init__(self)
        self.main = parent
        self.setFlags(QtGui.QGraphicsItem.ItemIgnoresTransformations)
        self.setPen(pen)
        self.setBrush(brush)
        self.path = QtGui.QPainterPath()
        self.path.moveTo(QtCore.QPointF(-8, -8))
        self.path.lineTo(QtCore.QPointF(8, 8))
        self.path.moveTo(QtCore.QPointF(-8, 8))
        self.path.lineTo(QtCore.QPointF(8, -8))
        self.font = font
        self.text = text
        if text != '':
            self.setText(pen, brush, font, text)
        self.setPath(self.path)

    def chooseText(self, pen, brush, font):
        """
        pour choisir le texte à la création du point.
        """
        try:
            proposedText = POINT_NAMES[0]
        except:
            proposedText = ''
        text, ok = QtGui.QInputDialog.getText(self.main, utils.PROGLABEL,
                                              QtGui.QApplication.translate('main', 'Point Name:'),
                                              QtGui.QLineEdit.Normal,
                                              proposedText)
        if ok and text != '':
            self.text = text
            changePointNames(text)
            self.textItem = QtGui.QGraphicsTextItem(text, self)
            self.textItem.setFont(font)
            self.textItem.setDefaultTextColor(pen.color())
            self.textItem.setPos(QtCore.QPointF(8, -32))
        return ok

    def setText(self, pen, brush, font, text):
        """
        Pour définir le texte (à l'ouverture d'un fichier par exemple)
        """
        self.text = text
        self.textItem = QtGui.QGraphicsTextItem(text, self)
        self.textItem.setFont(font)
        self.textItem.setDefaultTextColor(pen.color())
        self.textItem.setPos(QtCore.QPointF(8, -32))  



